import torch
import random


def isint(s):
    try:
        int(s)
        return True
    except ValueError:
        return False


def readchardict(charfile):
    chardict = {"OOV": 1}
    with open(charfile, "r", encoding="utf-8") as f:
        i = 2
        for line in f:
            chardict[line[0]] = i
            i += 1
    return chardict


def readpostfile(fname):
    tags2int = {}
    int2tags = {}
    with open(fname, "r", encoding="utf-8") as f:
        i = 0
        for line in f:
            if len(line)>0:
                tag = line.rstrip()
                tags2int[tag] = i
                int2tags[i] = tag
                i += 1
    return tags2int, int2tags


def prepareforpostagging(sentences):
    new_sentences = []
    for s in sentences:
        ns = [[w["FORM"], w["UPOS"]] for w in s if "COMMENT" not in w and isint(w["ID"])]
        new_sentences.append(ns)
    return new_sentences


def getcharlvltensors(sentences, chardict, postags, device=None):
    words = []
    tags = []
    for s in sentences:
        ws = []
        ts = []
        for w in s:
            cl = []
            for c in w[0]:
                if c in chardict:
                    cl.append(chardict[c])
                else:
                    cl.append(chardict["OOV"])
            ws.append(torch.tensor(cl, device=device).view(1, -1))
            ts.append(postags[w[1]])
        words.append(ws)
        tags.append(torch.tensor(ts, device=None).view(1, -1))
    return words, tags


def getsentencebatch(sentences, chardict, postags, sentencenr=1, device=None):
    batches = []
    targets = []
    nb = []
    nt = []
    i = 0
    for s in sentences:
        for w in s:
            cl = []
            for c in w[0]:
                if c in chardict:
                    cl.append(chardict[c])
                else:
                    cl.append(chardict["OOV"])
            nb.append(torch.tensor(cl, device=device).view(1, -1))
            nt.append(postags[w[1]])
        i += 1
        if i%sentencenr == 0:
            batches.append(nb)
            targets.append(torch.tensor(nt, device=device).view(1, -1))
            nb = []
            nt = []
            i = 0
    return batches, targets


def getsamelengthbatches(sentences, chardict, postags, length, device=None):
    wordlist = []
    targetlist = []
    for s in sentences:
        for w in s:
            cl = []
            for c in w[0]:
                if c in chardict:
                    cl.append(chardict[c])
                else:
                    cl.append(chardict["OOV"])
            wordlist.append(torch.tensor(cl, device=device).view(1, -1))
            targetlist.append(postags[w[1]])
    leftover = len(wordlist) % length
    wordlist = wordlist[:-leftover]
    targetlist = targetlist[:-leftover]
    batches = [wordlist[i:i+length] for i in range(0, len(wordlist), length)]
    targets = [torch.tensor(targetlist[i:i+length], device=device).view(1, -1) for i in range(0, len(targetlist), length)]
    return batches, targets


def shuffle_batches(batches, targets, seed=None):
    m = len(batches)
    bl = batches
    tl = targets
    n = 0
    if seed is None:
        rand = random.Random()
    else:
        rand = random.Random(seed)
    for i in range(0, m):
        j = rand.randrange(m)
        tempb = bl[i]
        bl[i] = bl[j]
        bl[j] = tempb
        tempt = tl[i]
        tl[i] = tl[j]
        tl[j] = tempt
    return bl, tl


class StatObject():
    def __init__(self, labels):
        self.labels = labels
        self.matrix = [[0 for i in range(0, len(self.labels))] for j in range(0, len(self.labels))]
        self.total_true = [0 for i in range(0, len(self.labels))]
        self.true_positives = [0 for i in range(0, len(self.labels))]
        self.recall = [0 for i in range(0, len(self.labels))]
        self.precision = [0 for i in range(0, len(self.labels))]
        self.total_predicted = [0 for i in range(0, len(self.labels))]
        self.accuracy = -1
        self.total = 0
        self.needs_eval = True


    def add_prediction(self, true_label, prediction):
        self.matrix[true_label][prediction] += 1
        self.needs_eval = True


    def evaluate(self):
        if self.needs_eval:
            self.total_true = [0 for i in range(0, len(self.labels))]
            self.true_positives = [0 for i in range(0, len(self.labels))]
            self.total_predicted = [0 for i in range(0, len(self.labels))]
            for i in range(0, len(self.matrix)):
                for j in range(0, len(self.matrix[i])):
                    self.total_true[i] += self.matrix[i][j]
                    self.total_predicted[j] += self.matrix[i][j]
                    self.total += self.matrix[i][j]
                    if i==j:
                        self.true_positives[i] += self.matrix[i][j]
            self.accuracy = sum(self.true_positives)/self.total if self.total > 0 else -1
            for i in range(0, len(self.true_positives)):
                self.precision[i] = self.true_positives[i]/self.total_predicted[i] if self.total_predicted[i]>0 else -1
                self.recall[i] = self.true_positives[i]/self.total_true[i] if self.total_true[i]>0 else -1
            self.needs_eval = False


    def clear(self):
        self.matrix = [[0 for i in range(0, len(self.labels))] for j in range(0, len(self.labels))]
        self.total_true = [0 for i in range(0, len(self.labels))]
        self.true_positives = [0 for i in range(0, len(self.labels))]
        self.recall = [0 for i in range(0, len(self.labels))]
        self.precision = [0 for i in range(0, len(self.labels))]
        self.total_predicted = [0 for i in range(0, len(self.labels))]
        self.accuracy = -1
        self.total = 0
        self.needs_eval = True


    def __str__(self):
        self.evaluate()
        s = "Row: Actual labels, Column: Predicted labels\n"
        s += "     "
        for l in self.labels:
            s += "|{:^5}".format(l)
        s += "|Total\n"
        for i in range(0, len(self.labels)):
            s += "{:^5}".format(self.labels[i])
            for el in self.matrix[i]:
                s += "|{:^5}".format(el)
            s += "|{:^5}\n".format(self.total_true[i])
        s += "Total"
        for i in range(0, len(self.labels)):
            s += "|{:^5}".format(self.total_predicted[i])
        s += "|{:^5}\n\n".format(self.total)
        s += "Precision\n"
        for l in self.labels:
            s += "|{:^5}".format(l)
        s += "|\n"
        for i in range(0, len(self.labels)):
            s += "|{:^2.3f}".format(self.precision[i])
        s += "|\n"
        s += "Recall\n"
        for l in self.labels:
            s += "|{:^5}".format(l)
        s += "|\n"
        for i in range(0, len(self.labels)):
            s += "|{:^2.3f}".format(self.recall[i])
        s += "|\n"
        s += "Accuracy: {:.5f}\n".format(self.accuracy)
        return s
